# Spots removal mask maker Ver. 0.1
# revised 2024.5.19
# (c) 2024 Ohnishi, Yasuo 

from ij import IJ
from ij import WindowManager as WM
from ij.io import OpenDialog
from loci.plugins import BF
from java.awt.event import AdjustmentListener, ItemListener, TextListener
from ij.plugin import ChannelSplitter, Duplicator
from ij import plugin
from ij.process import ImageConverter
from ij.gui import GenericDialog
import os
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')

#
class imagePreviewer(AdjustmentListener, ItemListener, TextListener):
    def __init__(self, imp, impWn, nfield_sigma, nfield_lowerTh, \
                                nfield_upperTh, nfield_factor):
        #imp: RGB preview image, impWn: imp for Image preview window
        #nfield_sigma: Sigma for Gaussian blur
        #nfield_lowerTh: lower threshold, nfield_upperTh: uppper threshold
        #nfield_factor: transparency factor
        self.imp = imp
        self.impWn = impWn
        self.impOrg = Duplicator().run(imp) #
        self.impPreviewMask = Duplicator().run(imp)
        self.impPreviewBlur = Duplicator().run(imp)
        if LnProcess == True:
            impWn.getProcessor().gamma(1.0 / 2.2) #L to P
        self.impWn.show() #
        WM.getActiveWindow().toFront()
        self.nfield_sigma = nfield_sigma
        self.nfield_lowerTh = nfield_lowerTh
        self.nfield_upperTh = nfield_upperTh
        self.nfield_factor = nfield_factor
        self.sigma = float(nfield_sigma.getText())
        self.lowerTh = float(nfield_lowerTh.getText()) 
        self.upperTh = float(nfield_upperTh.getText()) 
        self.factor = float(nfield_factor.getText()) 

    def adjustmentValueChanged(self, event):

        if (IJ.isMacOSX() == False) and event.getValueIsAdjusting():
            return
        sigmaTmp = float(self.nfield_sigma.getText())
        lowerThTmp = float(self.nfield_lowerTh.getText())
        upperThTmp = float(self.nfield_upperTh.getText())
        factorTmp = float(self.nfield_factor.getText())
        if (upperThTmp <= lowerThTmp):
            self.nfield_upperTh.setText(str(lowerThTmp))
            upperThTmp = lowerThTmp
        elif (lowerThTmp >= upperThTmp):
            self.nfield_lowerTh.setText(str(upperThTmp))
            lowerThTmp = upperThTmp

        if (self.sigma != sigmaTmp):
            self.sigma = sigmaTmp
            self.execBlur()
        elif (self.lowerTh != lowerThTmp):
            self.lowerTh = lowerThTmp
            self.makePrvMask()
        elif (self.upperTh != upperThTmp):
            self.upperTh = upperThTmp
            self.makePrvMask()
        elif (self.factor != factorTmp):
            self.factor = factorTmp
            self.makePrvMask()


    def execBlur(self):        #

        impTmpBlur = Duplicator().run(self.imp) #
        impTmpBlur.getProcessor().blurGaussian(self.sigma)
        ip = impTmpBlur.getProcessor().duplicate()
        ip2 = impTmpBlur.getProcessor().duplicate()
        if LnProcess == True:
            ip2.gamma(1.0 / 2.2) #L
        self.impPreviewBlur.setProcessor(ip)
        self.impPreviewBlur.updateAndDraw()
        self.impWn.setProcessor(ip2)
        self.impWn.updateAndDraw()
        WM.getActiveWindow().toFront()

    def makePrvMask(self):        #

        if LnProcess == True:
            impPreviewMaskTmp = makeMask(self.imp, self.impPreviewBlur, \
                        P2L(self.lowerTh), P2L(self.upperTh) ,self.factor, 8)
        else:
            impPreviewMaskTmp = makeMask(self.imp, self.impPreviewBlur, \
                                    self.lowerTh, self.upperTh ,self.factor, 8)

        ip = impPreviewMaskTmp.getProcessor().duplicate()
        ip2 = impPreviewMaskTmp.getProcessor().duplicate()
        # l to p
        if LnProcess == True:
            ip2.gamma(1.0 / 2.2) #L
        #
        self.impPreviewMask.setProcessor(ip)
        self.impPreviewMask.updateAndDraw()
        self.impWn.setProcessor(ip2)
        self.impWn.updateAndDraw()
        WM.getActiveWindow().toFront()

#---------------------
def makeMask(imp, impBlur, lowerTh, upperTh, factor, BitDepth):
    ic = plugin.ImageCalculator()
    impMaskTmp = \
        ic.run("Subtract create", impBlur, imp)
    impMaskTmp.getProcessor().subtract(((2 ** BitDepth) - 1) * (lowerTh / 100))
    scale = 100.0 / upperTh
    impMaskTmp.getProcessor().multiply(scale)
    impMaskTmp.getProcessor().multiply(factor)

    return impMaskTmp

#---------------------
def parameterUI(impPrv, colorCh):
    #impPrv: RGB preview image, impWn: imp for Image preview window
    #nfield_sigma: Sigma for Gaussian blur
    #nfield_lowerTh: lower threshold, nfield_upperTh: uppper threshold
    #nfield_factor: transparency factor    
    impPrv.show()
    WM.getActiveWindow().toFront()
    gd = GenericDialog("Spots Remover / Parameter Setting")

    gd.addSlider("Sigma of Gaussian blur", 0.0, 100.0, 20.0, 0.01)   
    # Slider 0 & NField 0
    gd.addSlider("Mask: lower threshold", 0.0, 100.0, 0.0, 0.01)   
    # Slider 1 & NField 1
    gd.addSlider("Mask: upper threshold", 0.0, 100.0, 100.0, 0.01)   
    # Slider 2 & NField 2
    gd.addSlider("Effect factor for mask", 0.0, 5.0, 1.0, 0.01)   
    # Slider 3 & NField 3

    slider_sigma  = gd.getSliders().get(0) # slider 0
    slider_lowerTh  = gd.getSliders().get(1) # slider 1
    slider_upperTh  = gd.getSliders().get(2) # slider 1
    slider_factor  = gd.getSliders().get(3) # slider 1
    nfield_sigma = gd.getNumericFields().get(0) # NField 0
    nfield_lowerTh = gd.getNumericFields().get(1) # NField 1
    nfield_upperTh = gd.getNumericFields().get(2) # NField 1
    nfield_factor = gd.getNumericFields().get(3) # NField 1

    impWn = impPrv.duplicate() # imp as preview window
    previewer = imagePreviewer(impPrv, impWn, \
        nfield_sigma, nfield_lowerTh, nfield_upperTh, nfield_factor)

    slider_sigma.addAdjustmentListener(previewer) # listener for slider #0
    slider_lowerTh.addAdjustmentListener(previewer) # listener for slider #1
    slider_upperTh.addAdjustmentListener(previewer) # listener for slider #2
    slider_factor.addAdjustmentListener(previewer) # listener for slider #3

    gd.showDialog()

    sigma =  gd.getNextNumber() # NField 0
    lowerTh =  gd.getNextNumber() # NField 1
    upperTh =  gd.getNextNumber() # NField 2
    factor =  gd.getNextNumber() # NField 3

    slider_sigma.removeAdjustmentListener(previewer) # rmv listener for sigma
    slider_lowerTh.removeAdjustmentListener(previewer) # rmv listener for lower
    slider_upperTh.removeAdjustmentListener(previewer) # rmv listener for upper
    slider_factor.removeAdjustmentListener(previewer) # rmv listener for factor
    impWn.hide()
    impWn.close()
    return sigma, lowerTh, upperTh, factor

#-----------------
def convertToLuminosity(impOrg, weighted):
    # convert RGB 3 color image to mono Luminosity image
    # impOrg: original imagePlus
    # wighted: use Luminosity (perceptually weighted grayscale)
    #                                       instead of lightness
    imp = Duplicator().run(impOrg)
    impChs = ChannelSplitter.split(imp)
    for i in impChs:
        i.getProcessor().setLut(None)
    ic = plugin.ImageCalculator()
    if weighted == True:
        impChs[0].getProcessor().multiply(0.299)
        impChs[1].getProcessor().multiply(0.587)
        impChs[2].getProcessor().multiply(0.114)
    else:
        impChs[0].getProcessor().multiply(0.3333)
        impChs[1].getProcessor().multiply(0.3334)
        impChs[2].getProcessor().multiply(0.3333)
    impTmp1 = ic.run("Add created", impChs[0], impChs[1])
    impLuminosity = ic.run("Add created", impTmp1, impChs[2])
    impTmp1.close()
    imp.close()
    for i in impChs:
        i.close()
    return impLuminosity
#----------------------------------
#
class selectImagePreviewer(ItemListener):
    def __init__(self, imp, impPreview, impStack, choice_channel):
        #imp: Original image, impPreview: imp for Image preview window
        #impStack: separated channnel image imp array
        #choice_channel: Selector for channnel
        self.imp = imp
        self.impPreview = impPreview
        ImageConverter(self.imp).convertToRGB()
        self.impStack = impStack
        self.impPreview.show() #
        WM.getActiveWindow().toFront()
        self.choice_channel = choice_channel

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        choice_item  = self.choice_channel.getSelectedItem()
        if choice_item == "Red": 
            ipr = self.impStack[0].getProcessor().duplicate()
            self.impPreview.setProcessor(ipr) #
            self.impPreview.updateAndDraw()
            WM.getActiveWindow().toFront()
        elif choice_item == "Green": 
            ipg = self.impStack[1].getProcessor().duplicate()
            self.impPreview.setProcessor(ipg) #
            WM.getActiveWindow().toFront()
        elif choice_item == "Blue": 
            ipb = self.impStack[2].getProcessor().duplicate()
            self.impPreview.setProcessor(ipb) #
            WM.getActiveWindow().toFront()
        elif choice_item == "Lightness": 
            impl = convertToLuminosity(self.imp, False)
            ipl = impl.getProcessor().duplicate()
            self.impPreview.setProcessor(ipl) #
            WM.getActiveWindow().toFront()
        elif choice_item == "Luminosity": 
            impl = convertToLuminosity(self.imp, True)
            ipl = impl.getProcessor().duplicate()
            self.impPreview.setProcessor(ipl) #
            WM.getActiveWindow().toFront()
        self.impPreview.updateAndDraw()
        self.impPreview.show()
        
#---------------------
def selectChannelUI(impPrv, StackSize):
    # dialog of selecting target channel for mask upper & lower threshold
    #impPrv: Original image for preview
    #impWn: imp for preview image window
    #StackSize: stacksize of impPrv
    #impStack: separated channnel image imp array
    #choice_channel: Selector for channnel
    gd = GenericDialog("Spots Remover: Channel / film type Selection")
    # scolorCh = ["Red","Green","Blue"]
    scolorCh = ["Red","Green","Blue","Lightness", "Luminosity"]
    if StackSize == 3:
        impStack = ChannelSplitter.split(impPrv)
        for i in impStack:
            i.getProcessor().setLut(None)
        gd.addChoice("Select Target Channel for Mask", scolorCh, scolorCh[0])
    gd.addCheckbox("Negative film image", False) # CBox 0
    if StackSize == 3:
        choice_channel = gd.getChoices().get(0)
        impWn = impStack[0].duplicate() # imp as preview window
        previewerS = \
            selectImagePreviewer(impPrv, impWn, impStack, choice_channel)

        choice_channel.addItemListener(previewerS) #set listener for slider #0


    gd.showDialog()

    if StackSize == 3:
        colorCh =  gd.getNextChoice()
    else:
        colorCh = "Lightness"
    negaFilm = gd.getNextBoolean()
    if gd.wasCanceled():
        colorCh =  "None"
    if StackSize == 3:
        for i in impStack:
            i.hide()
            i.close()
        impWn.hide()
        impWn.close()
    return colorCh, negaFilm
#------------------------
def P2L(x):
    lx = (x  / 100) ** 2.2
    lx = lx * 100
    return lx

def L2P(x):
    px = (x / 100) ** (1 / 2.2)
    px = px * 100
    return px


#---Convert Value in 8, 16, 32bit mode
def convValue(Value, BitDepth, imp):
#     Converting Luminance Value each in 8, 16, 32bit mode
        if BitDepth == 16:
            Value = Value * 256
        elif BitDepth == 32:
            L_max = set_Lmax(imp)
            Value = Value * (L_max / 255.999)
        else: # 8 or 24
            Value = Value
        return Value
#-----------------
def set_Lmax(imp):
    # set Maximum value of Lightness
    L_max = imp.getProcessor().getMax()
    if int(L_max) > 65536:
        L_max = 16777215
    elif int(L_max) > 256:
        L_max = 65535
    elif int(L_max) > 1.0:
        L_max = 255
    else:
        L_max = 1.0
    return L_max

#--- image type detection
def imageTypeDetection(imp):
    StackSize = imp.getStackSize()
    BitDepth = imp.getBitDepth()
    return StackSize, BitDepth

#----load image-------
def load_image():
    imgCurrent = WM.getCurrentImage()
    if not imgCurrent:
        od = OpenDialog("Choose a file")
        folder = od.getDirectory()
        filename = od.getFileName()
    # Always Open file with Bio-format Plugin
        if filename:
            imps = BF.openImagePlus(folder + filename)
            imp0 = imps[0] #imp0: Original Image
            imp0.show()
        else:
            IJ.showMessage("File is not selected.")
    else: # target image is already open
        imp0 = imgCurrent
        filename = imp0.getTitle()
        try:
            folder = imp0.getOriginalFileInfo().directory
        except:
            folder = IJ.getDirectory("Output_directory")
    return imp0, filename, folder


#---- Main ----
LnProcess = True # Do internally process imege in Linear? (True/False)
imp0, filename, folder = load_image()
StackSize, BitDepth = imageTypeDetection(imp0)
if BitDepth ==24: # if imp0 is color processor
    stackSize =3
    BitDepth ==8
if filename:
    IJ.log("folder: " + str(folder))
    IJ.log("filename: " + str(filename))
    IJ.log("StackSize: " + str(StackSize))
    IJ.log("BitDepth: " + str(BitDepth))

impPrv = Duplicator().run(imp0) # imp1: preview imp
# call select processing channel UI
##### UI channel selector #####
colorCh, negaFilm = selectChannelUI(impPrv, StackSize)
# colorCh: selected channel name
if negaFilm == True: # If original is negative film > invert imp
    impPrv.getProcessor().invert()
    imp0.getProcessor().invert()
#######
if StackSize ==3: # color image
    impprvsCh = ChannelSplitter.split(impPrv)
    impCh = ChannelSplitter.split(imp0)
    for i in impprvsCh:
        i.setLut(None)
        ImageConverter(i).convertToGray8()
        BD = i.getBitDepth()

    # C0: name of color ch. C1: name of inverted color ch.
    # impPrv1: imp of preview of the ch.

if StackSize ==3: # color image
    if colorCh == "Blue":
        impPrv1 = Duplicator().run(impprvsCh[2])
        imp1 = Duplicator().run(impCh[2])
    elif colorCh == "Red":
        impPrv1 = Duplicator().run(impprvsCh[0])
        imp1 = Duplicator().run(impCh[0])
    elif colorCh == "Green":
        impPrv1 = Duplicator().run(impprvsCh[1])
        imp1 = Duplicator().run(impCh[1])
    elif colorCh == "Luminosity":
        impPrv0 = Duplicator().run(impPrv)
        ImageConverter(impPrv0).convertToRGB()
        impPrv1 = convertToLuminosity(impPrv0, True)
        impPrv0.close()
        imp1 = convertToLuminosity(imp0, True)
    elif colorCh == "Lightness":
        impPrv0 = Duplicator().run(impPrv)
        ImageConverter(impPrv0).convertToRGB()
        impPrv1 = convertToLuminosity(impPrv0, False)
        impPrv0.close()
        imp1 = convertToLuminosity(imp0, False)
elif colorCh != "None": # monochrome image
    ImageConverter(impPrv).convertToGray8()
    impPrv1 = impPrv
    imp1 = Duplicator().run(imp0)
if colorCh != "None":
    impPrv1.getProcessor().setLut(None)
    imp1.getProcessor().setLut(None)
#   impPrv1: imp for preview image, imp1: imp for making mask 
    # p to l
    if LnProcess == True: # convert to linear image
        impPrv1.getProcessor().gamma(2.2)
        imp1.getProcessor().gamma(2.2)
    ##### UI parameter #####
    sigma, lowerTh, upperTh, factor = parameterUI(impPrv1, colorCh)
    impBlur1 = Duplicator().run(imp1)
    impBlur1.getProcessor().blurGaussian(sigma)
    print ("lowerTh upperTh: ", str(lowerTh), str(upperTh))
    # P to L
    if LnProcess == True:
        lowerTh = P2L(lowerTh)
        upperTh = P2L(upperTh)
    #
    finalmask = makeMask(imp1, impBlur1, lowerTh, upperTh, factor, BitDepth)
    if LnProcess == True:
        impBlur1.getProcessor().gamma(1.0 / 2.2)
        finalmask.getProcessor().gamma(1.0 / 2.2)
    IJ.saveAsTiff(finalmask, folder + filename + "_spots_remove_mask.tif")  
    IJ.showMessage("Process finished")
    imp0.hide()
    impPrv1.hide()
    
    imp0.close()
    imp1.close()
    impPrv.close()
    impPrv1.close()
    impBlur1.close()
    finalmask.close()

